#!/usr/bin/python

__author__ =  'Mohamed A. Bamakhrama'
__description__ = 'Dot Handler'

import os.path

class DotHandler:
	"""Dot Handler"""

	colors = ["red", "lightblue", "green", "yellow", "orange", "lightgrey", "yellow"]
	
	def __init__(self, pg_instance, output_dir, partitions):
		self.pg_instance = pg_instance
		self.output_dir = output_dir
		self.partitions = partitions
		self.num_of_apps = self.pg_instance.get_num_of_graphs()
	
	def generate_mapping_file(self):
		mapping = os.path.join(self.output_dir, "mapping.dot")
		fp = open(mapping, 'w')
		fp.write("digraph mapping {\n")
		fp.write("rankdir=TB;\n")
		
		partitions = self.partitions
		proc_cnt = 0
		task_cnt = 0
		for p in partitions:
			fp.write("subgraph cluster" + str(proc_cnt) + " {\n")
			u_sum = 0.0
			for t in p:
				if self.num_of_apps > 1:
					fp.write('\t' + t.get_full_name() + "_" + t.get_function() + " [ label = \"" + t.get_full_name() + " - U = " + str(t.get_utilization()) + "\"];\n")
				else:
					fp.write('\t' + t.get_name() + "_" + t.get_function() + " [ label = \"" + t.get_name() + " - U = " + str(t.get_utilization()) + "\"];\n")
									
				task_cnt += 1
				u_sum += t.get_utilization()
			fp.write('\tlabel=\"Processor' + str(proc_cnt) + ' - U = ' + str(u_sum) + '\";\n')
			proc_cnt += 1
			fp.write("}\n")

		# Done with processors. Print apps edges

		app_cnt = 0
		for g in self.pg_instance.get_graphs():
			levels = g.find_levels()
			for s in levels:
				for a in s:
					actor = g.get_actor(a.get_actor_id())
					for outport in actor.get_outports():
						ch = g.get_channel(outport)
						successor = g.get_actor(ch.get_destination())
						if self.num_of_apps > 1:
							fp.write(g.get_graph_name() + "_" + actor.get_name() + "_" + actor.get_function() + "[style=filled,color=\"" + self.colors[app_cnt % len(self.colors)] + "\"];\n")
							fp.write(g.get_graph_name() + "_" + successor.get_name() + "_" + successor.get_function() + "[style=filled,color=\"" + self.colors[app_cnt % len(self.colors)] + "\"];\n")
							fp.write(g.get_graph_name() + "_" + actor.get_name() + "_" + actor.get_function() + " -> " + g.get_graph_name() + "_" + successor.get_name() + "_" + successor.get_function() + ";\n")
						else:
							fp.write(actor.get_name() + "_" + actor.get_function() + "[style=filled,color=\"" + self.colors[app_cnt % len(self.colors)] + "\"];\n")
							fp.write(successor.get_name() + "_" + successor.get_function() + "[style=filled,color=\"" + self.colors[app_cnt % len(self.colors)] + "\"];\n")
							fp.write(actor.get_name() + "_" + actor.get_function() + " -> " + successor.get_name() + "_" + successor.get_function() + ";\n")
			app_cnt += 1
					
			
		fp.write('label=\"Mapping\";\n')
		fp.write("}\n")
		fp.flush()
		fp.close()

