#!/usr/bin/python

__author__ =  'Mohamed A. Bamakhrama'
__description__ = 'ESPAM Handler'

import os.path
from operator import attrgetter

def ML605(output_dir, m):
	platform_file = os.path.join(output_dir, "platform.pla")
	fp = open(platform_file, 'w')
	fp.write("""<?xml version="1.0" standalone="no"?>
<!DOCTYPE platform PUBLIC "-//LIACS//DTD ESPAM 1//EN" "http://www.liacs.nl/~cserc/dtd/espam_1.dtd">
<platform name="myPlatform">

""")
	for i in range(m):
		fp.write("""
\t<processor name="mb_""" + str(i) + """" type="MB" data_memory="131072" program_memory="131072">
\t\t<port name="IO_0" />
\t</processor>
""")

	fp.write("\n")
	fp.write("""\t<network name="CS" type="AXICrossbarSwitch">\n""")
	for i in range(m):
		fp.write("""\t\t<port name="IO_""" + str(i) + """" />\n""")
	fp.write("\t</network>\n")
	fp.write("\n")
	fp.write("\t<host_interface name=\"HOST_IF\" type=\"ML605\">\n")
	fp.write("\t</host_interface>\n")		
	fp.write("\n")
	for i in range(m):
		fp.write("""
\t<link name="BUS""" + str(i) + """">
\t\t<resource name="mb_""" + str(i) + """" port="IO_0" />
\t\t<resource name="CS" port="IO_""" + str(i) + """" />
\t</link>
""")
	fp.write("\n")		

	fp.write("</platform>")
	fp.close()


class EspamHandler:
	"""ESPAM Handler"""
	
	scheduler_labels = {	'PEDF-FF' : "dynamic-edf", 
				'PEDF-WF' : "dynamic-edf",
  		 		'RM-FF-LB' : "dynamic-freertos",
  		 		'RM-FF-EB' : "dynamic-freertos",
				'RM-LCE-LB' : "dynamic-freertos",
				'RM-LCE-EB' : "dynamic-freertos",
				'RM-LCR-LB' : "dynamic-freertos",
				'RM-LCR-EB' : "dynamic-freertos",
				'RM-LCG-LB' : "dynamic-freertos",
				'RM-LCG-EB' : "dynamic-freertos",
  			}
	
	supported_boards = {'ML605' : ML605}

	def __init__(self, pg_instance, board_type, output_dir, partitions):
		self.pg_instance = pg_instance
		self.num_of_apps = pg_instance.get_num_of_graphs()
		self.partitions = partitions
		self.num_of_processors = len(partitions)
		self.output_dir = output_dir
		self.board_handler = self.supported_boards[board_type]
		
		
	def generate_platform_file(self):
		self.board_handler(self.output_dir, self.num_of_processors)
		
		
	def generate_mapping_file(self):
		mapping_file = os.path.join(self.output_dir, "mapping.map")
		fp = open(mapping_file, 'w')
		fp.write("""<?xml version="1.0" standalone="no"?>
<!DOCTYPE mapping PUBLIC "-//LIACS//DTD ESPAM 1//EN" "http://www.liacs.nl/~cserc/dtd/espam_1.dtd">
<mapping name="myMapping">
""")
		partitions = self.partitions
		proc_cnt = 0
		for p in partitions:
			fp.write('\t<processor name="mb_' + str(proc_cnt) + '" scheduleType="' + self.scheduler_labels[self.pg_instance.get_scheduler_type()] + '">\n')
			proc_cnt += 1
			sorted_tasks = sorted(p, key = attrgetter('period'), reverse = True)
			priority = 1
			for t in sorted_tasks:
				t.set_priority(priority)
				priority = priority + 1
				name = t.get_name()
				if self.num_of_apps > 1:
					name = t.get_full_name()
				if self.pg_instance.get_scheduler_type().startswith("RM"):
					fp.write('\t\t<process name="' + name + '" period="' + str(t.get_period()) + '" priority="' + str(t.get_priority()) + '"/>\n')
				else:
					fp.write('\t\t<process name="' + name + '" execution="' + str(t.get_wcet()) + '" period="' + str(t.get_period()) + '" deadline="' + str(t.get_deadline()) + '" startTime="' + str(t.get_start_time()) + '" priority="' + str(t.get_priority()) + '"/>\n')
			fp.write("\t</processor>\n")
		fp.write("\n")
		graphs = self.pg_instance.get_graphs()
		for g in graphs:
			channels = g.get_channels()
			for c in channels:
				if self.num_of_apps > 1:
					fp.write('\t<fifo name="' + g.get_graph_name() + "_" + c.get_name() + '" size="' + str(c.get_buffer_size()) + '" />\n')
				else:
					fp.write('\t<fifo name="' + c.get_name() + '" size="' + str(c.get_buffer_size()) + '" />\n')
		fp.write("\n")
		fp.write("</mapping>")
		fp.close()

		
