#!/usr/bin/python

__author__ =  'Mohamed A. Bamakhrama'
__description__ = 'Mapping Finder'

import sys
import random
from operator import attrgetter

import Utilities
from Partition import Partition

class Mapping:

	LOWER_BOUND = 0
	EXACT_BOUND = 1
	
	RM_LOWER_BOUND = 0.65
	EDF_BOUND = 1.0


	@staticmethod
	def feasible_EDF(mapping, tasks):
		parts = [set() for i in range(len(mapping))]
		proc_cnt = 0
		for s in mapping:
			ts = Partition(Mapping.EDF_BOUND) 
			for t in s:
				task = tasks[t]
				f = ts.add_task(task)
				if not f:
					return None
			parts[proc_cnt] = ts.get_tasks()
			proc_cnt += 1
		return parts
		

	@staticmethod
	def feasible_RM_LB(mapping, tasks):
		parts = [set() for i in range(len(mapping))]
		proc_cnt = 0
		for s in mapping:
			ts = Partition(Mapping.RM_LOWER_BOUND) 
			for t in s:
				task = tasks[t]
				f = ts.add_task(task)
				if not f:
					return None
			parts[proc_cnt] = ts.get_tasks()
			proc_cnt += 1
		return parts


	@staticmethod
	def feasible_RM_EB(mapping, tasks):
		parts = [set() for i in range(len(mapping))]
		proc_cnt = 0
		for s in mapping:
			n = float(len(s))
			bound = Utilities.RM_exact_bound(n)
			ts = Partition(bound) 
			for t in s:
				task = tasks[t]
				f = ts.add_task(task)
				if not f:
					return None
			parts[proc_cnt] = ts.get_tasks()
			proc_cnt += 1
		return parts


	@staticmethod
	def contention(mapping, pg):
		channels = pg.get_mapped_channels(mapping)
		gamma =[[0 for i in range(len(mapping))] for j in range(len(mapping))]
		for c in channels:
			gamma[c[0]][c[1]] = 1
		max_f = 0
		for r in gamma:
			s = sum(r)
			if s > max_f:
				max_f = s
		if max_f > 1:
			return max_f
		else:
			return 0


	@staticmethod
	def find_PEDF_FF(pg):
		m = pg.get_min_num_of_processors()
		partitions = None
		while partitions == None:
			partitions = [set() for i in range(m)]
			for t in pg.get_tasks():
				done = False
				for p in partitions:
					total_u = 0.0
					for e in p:
						total_u = total_u + e.get_utilization()
					if (total_u + t.get_utilization()) <= Mapping.EDF_BOUND:
						p.add(t)
						done = True
						break
				if not done:
					partitions = None
					break
			if partitions != None:
				xp = [set() for k in range(len(partitions))]
				ii = 0
				for pp in partitions:
					for tt in pp:
						xp[ii].add(tt.get_task_id())
					ii += 1
				pg.update_graphs(xp)
				parts = Mapping.feasible_EDF(xp, pg.get_tasks())
				if parts != None:
					partitions = parts
					break
			m = m + 1
		return partitions
        
        
	@staticmethod
	def find_PEDF_FFD(pg):
		m = pg.get_min_num_of_processors()
                
                # sort tasks according to utilization in descending order
		tasks = pg.get_tasks()
                #for i in range(0, len(tasks)-1):
                        #swap_test = False
                        #for j in range(0, len(tasks) - i - 1):
                                #if tasks[j].get_utilization() < tasks[j+1].get_utilization():
                                        #tasks[j], tasks[j + 1] = tasks[j + 1], tasks[j]  # swap
                                #swap_test = True
                        #if swap_test == False:
                                #break
		sorted_tasks = sorted(tasks, key = attrgetter('utilization'), reverse = True)
                
		partitions = None	
		while partitions == None:
			partitions = [set() for i in range(m)]
			for t in sorted_tasks:
				done = False
				for p in partitions:
					total_u = 0.0
					for e in p:
						total_u = total_u + e.get_utilization()
					if (total_u + t.get_utilization()) <= Mapping.EDF_BOUND:
						p.add(t)
						done = True
						break
				if not done:
					partitions = None
					break
			if partitions != None:
				xp = [set() for k in range(len(partitions))]
				ii = 0
				for pp in partitions:
					for tt in pp:
						xp[ii].add(tt.get_task_id())
					ii += 1
				pg.update_graphs(xp)
				parts = Mapping.feasible_EDF(xp, pg.get_tasks())
				if parts != None:
					partitions = parts
					break
			m = m + 1
		return partitions
        
        
	@staticmethod
	def find_PEDF_WF(pg):
		m = pg.get_min_num_of_processors()
		partitions = None
		while partitions == None:
			pre_partitions = [Partition(Mapping.EDF_BOUND) for i in range(m)]
			for t in pg.get_tasks():
				done = False
				# Sort the partitions according to utilization in ascending order
				# insert to the first element (one with least U)
				# if insert is not sucessful, return null
				sorted_partitions = sorted(pre_partitions, key = attrgetter('utilization'), reverse = True)			
				for i in range(0, len(sorted_partitions)):
					inserted = sorted_partitions[i].add_task(t)
					if inserted:
						done = True
						break
				if not done:
					pre_partitions = None
					break
			if pre_partitions != None:
				partitions = [set() for i in range(m)]
				for i in range(m):
					partitions[i] = pre_partitions[i].get_tasks()
			else:
				partitions = None

			if partitions != None:
				xp = [set() for k in range(len(partitions))]
				ii = 0
				for pp in partitions:
					for tt in pp:
						xp[ii].add(tt.get_task_id())
					ii += 1
				pg.update_graphs(xp)
				parts = Mapping.feasible_EDF(xp, pg.get_tasks())
				if parts != None:
					partitions = parts
					break
			m = m + 1
		return partitions
        
        # comment out because python3 does not allow mixed space and tabs for indentation
        #@staticmethod
        #def find_PEDF_WFD(pg):
                #m = pg.get_min_num_of_processors()
                
                ## sort according to utilization in descending order
                #tasks = pg.get_tasks()
                ##for i in range(0, len(tasks)-1):
                        ##swap_test = False
                        ##for j in range(0, len(tasks) - i - 1):
                                ##if tasks[j].get_utilization() < tasks[j+1].get_utilization():
                                        ##tasks[j], tasks[j + 1] = tasks[j + 1], tasks[j]  # swap
                                ##swap_test = True
                        ##if swap_test == False:
                                ##break
                #sorted_tasks = sorted(tasks, key = attrgetter('utilization'), reverse = True)
                
                #partitions = None
                #while partitions == None:
                        #pre_partitions = [Partition(Mapping.EDF_BOUND) for i in range(m)]
                        #for t in sorted_tasks:
                                #done = False
                                ## Sort the partitions according to utilization in ascending order (Teddy:??? shouldnt be descending order)
                                ## insert to the first element (one with least U)
                                ## if insert is not sucessful, return null
                                #sorted_partitions = sorted(pre_partitions, key = attrgetter('utilization'), reverse = True)                     
                                #for i in range(0, len(sorted_partitions)):
                                        #inserted = sorted_partitions[i].add_task(t)
                                        #if inserted:
                                                #done = True
                                                #break
                                #if not done:
                                        #pre_partitions = None
                                        #break
                        #if pre_partitions != None:
                                #partitions = [set() for i in range(m)]
                                #for i in range(m):
                                        #partitions[i] = pre_partitions[i].get_tasks()
                        #else:
                                #partitions = None

                        #if partitions != None:
                                #xp = [set() for k in range(len(partitions))]
                                #ii = 0
                                #for pp in partitions:
                                        #for tt in pp:
                                                #xp[ii].add(tt.get_task_id())
                                        #ii += 1
                                #pg.update_graphs(xp)
                                #parts = Mapping.feasible_EDF(xp, pg.get_tasks())
                                #if parts != None:
                                        #partitions = parts
                                        #break
                        #m = m + 1
                #return partitions
        
        # comment out because python3 does not allow mixed space and tabs for indentation
        #@staticmethod
        #def find_PEDF_BFD(pg):
                #m = pg.get_min_num_of_processors()
                
                ## sort all tasks according to utilization in descending order
                #tasks = pg.get_tasks()
                #sorted_tasks = sorted(tasks, key = attrgetter('utilization'), reverse = True)
                
                #partitions = None
                #while partitions == None:
                        #pre_partitions = [Partition(Mapping.EDF_BOUND) for i in range(m)]
                        #for t in sorted_tasks:
                                #done = False
                                ## Sort the partitions according to utilization in ascending order.
                                ## and insert to the first element (one with least U left)
                                ## if insert is not sucessful, return null
                                #sorted_partitions = sorted(pre_partitions, key = attrgetter('utilization'), reverse = False)
                                #for i in range(0, len(sorted_partitions)):
                                        #inserted = sorted_partitions[i].add_task(t)
                                        #if inserted:
                                                #done = True
                                                #break
                                #if not done:
                                        #pre_partitions = None
                                        #break
                        #if pre_partitions != None:
                                #partitions = [set() for i in range(m)]
                                #for i in range(m):
                                        #partitions[i] = pre_partitions[i].get_tasks()
                        #else:
                                #partitions = None

                        #if partitions != None:
                                #xp = [set() for k in range(len(partitions))]
                                #ii = 0
                                #for pp in partitions:
                                        #for tt in pp:
                                                #xp[ii].add(tt.get_task_id())
                                        #ii += 1
                                #pg.update_graphs(xp)
                                #parts = Mapping.feasible_EDF(xp, pg.get_tasks())
                                #if parts != None:
                                        #partitions = parts
                                        #break
                        #m = m + 1
                #return partitions
                
                
        # U: unfolding; E: exhaustive search;
        # @param:
        # pg: platformGenerator
        # n_pe_upper: explore the design space consisting of up to n_pe_upper processors
        # pdg_file_name: name of pdg file
        # @return: 
	#@staticmethod
	#def find_PEDF_UE(pg, n_pe_upper, app_name, output_dir):
		#m_opt = pg.get_min_num_of_processors()
		#assert n_pe_upper >= m_opt, 'Error: Given upper bound on the number of PEs has to be greater than minimum required one.'
		
		#source_file = app_name
		
		#shutil.rmtree(source_file + "_work")
		#os.mkdir(source_file + "_work")
		#shutil.copy(source_file + ".c", source_file + "_work/")
		#shutil.copy("impldata.xml", source_file + "_work/")
		#os.chdir(source_file + "_work")
		
		
		### assume that pn/pntools are in the $PATH
		### (call pdgtrans in pntools)
		#infile = source_file + ".c"
		#outfile = open(source_file + ".yaml", "w")
		#child = subprocess.check_call(["c2pdg", "-func", "main", infile], stdout=outfile)
		#outfile.close()
                        
            
		### TODO: exhaustive search for best unfolding factor of the reduced FM example		
		#for d0 in range(1, 102 + 1):
			#for d1 in range(1, 17 + 1):
				#for d2 in range(1, 17 + 1):
					#for d3 in range(1, 17 + 1):
						### iterate over all factors
						#node0_unfold = "--modulo-split --node 0 factors" + 1 + "," + d0
						#node1_unfold = "--modulo-split --node 1 factors" + 1 + "," + d1
						#node2_unfold = "--modulo-split --node 2 factors" + 1 + "," + d2
						#node3_unfold = "--modulo-split --node 3 factors" + 1 + "," + d3
						##node4_unfold = "--modulo-split --node 4 factors" + 1 + "," + d4
						##node5_unfold = "--modulo-split --node 5 factors" + 1 + "," + d5
						##node6_unfold = "--modulo-split --node 6 factors" + 1 + "," + d6

            
						## assuming that pntools is added to $PATH
						#infile = open(source_file + ".yaml", "r")
						#outfile = open(source_file + "_parti.yaml", "w")
						#child = subprocess.check_call(["pdgtrans " + node0_unfold + " " + node1_unfold + " " + node2_unfold + " " + node3_unfold], \
								#shell=True, stdin=infile, stdout=outfile)
						#infile.close(); outfile.close()
						
						#infile = open(source_file + ".yaml", "r")
						#outfile = open(source_file + "-pn.yaml", "w")
						#child = subprocess.check_call(["pn"], stdin=infile, stdout=outfile)
						#infile.close(); outfile.close()
						
						#infile = open(source_file + "_parti.yaml", "r")
						#outfile = open(source_file + "_parti-pn.yaml", "w")
						#child = subprocess.check_call(["pn"], stdin=infile, stdout=outfile)
						#infile.close(); outfile.close()
						
						#infile = open(source_file + "_parti-pn.yaml", "r")
						#outfile = open(source_file + "_parti-adg.yaml", "w")
						#child = subprocess.check_call(["pn2adg"], stdin=infile, stdout=outfile)
						#infile.close(); outfile.close()
						
						#infile = open(source_file + "-adg.yaml", "r")
						#outfile = open(source_file + ".gph", "w")
						#child = subprocess.check_call(["adg2csdf"], stdin=infile, stdout=outfile)
						#infile.close(); outfile.close()
						
						#infile = open(source_file + "_parti-adg.yaml", "r")
						#outfile = open(source_file + "_parti.gph", "w")
						#child = subprocess.check_call(["adg2csdf"], stdin=infile, stdout=outfile)
						#infile.close(); outfile.close()
						
						#infile = open(source_file + "_parti-adg.yaml", "r")
						#outfile = open(source_file + "_parti.xml", "w")
						#child = subprocess.check_call(["adg2csdf", "-3"], stdin=infile, stdout=outfile)
						#infile.close(); outfile.close()
						
						#infile = open(source_file + "_parti.gph", "r")
						#outfile = open(source_file + "_parti.csf", "w")
						#child = subprocess.check_call(["CSDFConverter.py"], stdin=infile, stdout=outfile)
						#infile.close(); outfile.close()
						
						#MAX_SCALING = 10000 # empiracal value: upper bound on scaling period
						#for s in range(1, MAX_SCALING):
							#outfile = open("output", "w")
							#child = subprocess.check_call(["PlatformGenerator.py " + os.getcwd() + " PEDF-FFD " + s + " 0 "+ source_file + "_parti.csf"], shell=True, stdout=outfile)
							#out_lines = outfile.readlines()
							## sink period
							
							## nr. of processors
							## total utilization
							
							
							#outfile.close()
							
                
                
                
        
	@staticmethod
	def find_RM_FF(pg, bound_type):
					
		m = pg.get_min_num_of_processors()
		partitions = None
		while partitions == None:
			partitions = [set() for i in range(m)]
			for t in pg.get_tasks():
				done = False
				for p in partitions:
					total_u = 0.0
					
					if bound_type == Mapping.LOWER_BOUND:
						bound = Mapping.RM_LOWER_BOUND
					elif bound_type == Mapping.EXACT_BOUND:
						bound = Utilities.RM_exact_bound(len(p) + 1)
					else:
						assert False
						
					for e in p:
						total_u = total_u + e.get_utilization()
					if (total_u + t.get_utilization()) <= bound or (t.get_utilization() > bound and len(p) == 0):
						p.add(t)
						
						done = True
						break
				if not done:
					partitions = None
					break
			if partitions != None:
				xp = [set() for k in range(len(partitions))]
				ii = 0
				for pp in partitions:
					for tt in pp:
						xp[ii].add(tt.get_task_id())
					ii += 1
				pg.update_graphs(xp)
				if bound_type == Mapping.LOWER_BOUND:
					parts = Mapping.feasible_RM_LB(xp, pg.get_tasks())
				elif bound_type == Mapping.EXACT_BOUND:
					parts = Mapping.feasible_RM_EB(xp, pg.get_tasks())
				else:
					assert False
				if parts != None:
					partitions = parts
					break
			m = m + 1
		return partitions


	@staticmethod
	def find_RM_LCE(pg, bound_type):
		# Source: http://compprog.wordpress.com/2007/10/15/generating-the-partitions-of-a-set/
		def generate_all(set_):
			if not set_:
				yield []
				return
			for i in xrange(2**len(set_)/2):
				parts = [set(), set()]
				for item in set_:
					parts[i&1].add(item)
					i >>= 1
				for b in generate_all(parts[1]):
					yield [parts[0]]+b

		if bound_type == Mapping.LOWER_BOUND:
			feasible = Mapping.feasible_RM_LB
		elif bound_type == Mapping.EXACT_BOUND:
			feasible = Mapping.feasible_RM_EB
		else:
			assert False

		n = pg.get_num_of_tasks()
		ts = set([i for i in range(n)])
		m = pg.get_min_num_of_processors()
		eligible = []
		max_cont = sys.maxint
		total_mappings = 0
		total_feasible = 0
		while m <= n:
			p = generate_all(ts)
			for x in p:
				if len(x) <= m:
					pg.update_graphs(x)
					partitions = feasible(x, pg.get_tasks())
					total_mappings += 1
					if partitions != None:
						total_feasible += 1
						c = Mapping.contention(partitions, pg)
						if c <= max_cont:
							e = (x, partitions, c)
							eligible.append(e)
							max_cont = c
			if eligible == []:
				m += 1
				total_mappings = 0
				total_feasible = 0
			else:
				break
		# filter out elements with cont > max_cont
		for e in eligible:
			if e[2] > max_cont:
				eligible.remove(e)
		el = []
		for e in eligible:
			el.append(e[1])
		sys.stdout.write("Exhaustive search found " + str(len(el)) + " best mappings out of " + str(total_feasible) + " total feasible mappings and " + str(total_mappings) + " total mappings\n")
		return el

	
	@staticmethod
	def find_RM_LCR(pg, bound_type):
		def generate_random(n, m, num_of_mappings):
			mappings = []
			while len(mappings) < num_of_mappings:
				part = [set() for i in range(m)]
				ts = set([i for i in range(n)])
				while len(ts) > 0:
					index = random.randint(0, n - 1)
					e = index
					if e in ts:
						pindex = random.randint(0, m - 1)
						part[pindex].add(e)
						ts.remove(e)
				if part in mappings:
					None
				else:
					mappings.append(part)
			return mappings
	
		if bound_type == Mapping.LOWER_BOUND:
			feasible = Mapping.feasible_RM_LB
		elif bound_type == Mapping.EXACT_BOUND:
			feasible = Mapping.feasible_RM_EB
		else:
			assert False
	
		n = pg.get_num_of_tasks()
		m = pg.get_min_num_of_processors()
		eligible = []
		max_cont = sys.maxint
		total_mappings = 0
		total_feasible = 0
		while m <= n:
			p = generate_random(n,m,200)
			for x in p:
				if len(x) <= m:
					pg.update_graphs(x)
					partitions = feasible(x, pg.get_tasks())
					total_mappings += 1
					if partitions != None:
						total_feasible += 1
						c = Mapping.contention(partitions, pg)
						if c <= max_cont:
							e = (x, partitions, c)
							eligible.append(e)
							max_cont = c
			if eligible == []:
				m += 1
				total_mappings = 0
				total_feasible = 0
			else:
				break
		# filter out elements with cont > max_cont
		for e in eligible:
			if e[2] > max_cont:
				eligible.remove(e)
		el = []
		for e in eligible:
			el.append(e[1])
		return el


	@staticmethod
	def find_RM_LCG(pg, bound_type):

		# Needed for the Genetic algorithm
		# pyevolve (http://pyevolve.sourceforge.net)
		from pyevolve import G1DList
		from pyevolve import GSimpleGA
		from pyevolve import Consts
			
		def eval_func(chromosome):
			score = 0.0
			ps = [set() for i in range(m)]
			i = 0
			for v in chromosome:
				ps[v].add(i)
				i += 1
			pg.update_graphs(ps)
			partitions = feasible(ps, pg.get_tasks())
			if partitions == None:
				return score
			else:
				score += len(chromosome)**2

			cont = Mapping.contention(partitions, pg)
			score -= cont
			return score

		if bound_type == Mapping.LOWER_BOUND:
			feasible = Mapping.feasible_RM_LB
		elif bound_type == Mapping.EXACT_BOUND:
			feasible = Mapping.feasible_RM_EB
		else:
			assert False
			
		n = pg.get_num_of_tasks()
		m = pg.get_min_num_of_processors() 
		score = 0.0

		while m <= n:
			genome = G1DList.G1DList(n)
			genome.evaluator.set(eval_func)
			genome.setParams(rangemin = 0, rangemax = (m - 1))
			ga = GSimpleGA.GSimpleGA(genome)
			ga.setMinimax(Consts.minimaxType["maximize"])
			ga.setPopulationSize(100)
			ga.setGenerations(100)
			ga.setCrossoverRate(0.9)
			ga.setMutationRate(0.05)
			ga.evolve(freq_stats = 10)
			best =  ga.bestIndividual()
			score = best.score
			if score > 0.0:
				break
			else:
				sys.stdout.write("No suitable mapping found using " + str(m)  + " processors\n")
				m += 1

		bm = [set() for i in range(m)]
		i = 0
		for v in best:
			bm[v].add(pg.get_tasks()[i])
			i += 1
		return bm
        
        
                
            

