csdf-rtschedtools 
=================

csdf-rtschedtools is a framework for hard-real-time schedulability analysis of streaming applications modeled as acyclic Cyclo-Static Dataflow (CSDF) graphs. Given a set of graphs, where each graph corresponds to an applications, it performs the following:

* Constructing a complete task-set of asynchronous periodic tasks for each application
* Computing the maximum throughput and minimum latency of each application under strictly periodic scheduling
* Deriving the minimum number of processors needed to schedule the set of applications on homogeneous systems by different scheduling algorithms (e.g., Partitioned EDF and optimal)
* Generating platform and mapping specifications which can be used as input to the Daedalus^RT design flow


Task-set Construction
---------------------
Given an acyclic CSDF graph, it derives an asynchronous set of periodic tasks. The deadline settings is customizable and ranges from implicit-deadline to constrained-deadline. The task-set derivation is implemented in ACSDFModel.py and can be invoked as follows:

    > ./ACSDFModel.py csf_file deadline_type [deadline_configuration]
where:

* `csf_file`: is the full path to the .csf file containing the CSDF graph. The .csf file format is explained below under "Graph Formats".
* `deadline_type`: can take one of three values: `--idp-sufficient`, `--idp-minimal`, and `--cdp-minimal`. `--idp-sufficient` refers to the implicit-deadline periodic (IDP) model with sufficient start times and buffer sizes. `--idp-minimal` refers to the IDP model with minimum start times and buffer sizes. `--cdp-minimal` refers to the constrained-deadline periodic (CDP) model with minimum start times and buffer sizes.
* `deadline_configuration`: This parameters is needed only when "deadline_type=--cdp-minimal". It refers to the value of the deadline. The supported values are integers in the range [1,5]. The interpretation is as follows:

    1. `D_i = T_i`
    2. `D_i = C_i + (T_i - C_i)*3/4`
    3. `D_i = C_i + (T_i - C_i)/2`
    4. `D_i = C_i + (T_i - C_i)/4`
    5. `D_i = C_i`

where `D_i` is the deadline of task `f_i`, `T_i` is the period of `f_i`, and `C_i` is the WCET of `f_i`

`ACSDFModel.py` produces its output to `stdout`. So it might be useful for the user to redirect the output to a file.

Computing the throughput and latency
------------------------------------
The minimum latency and maxmimum throughput are automatically computed during the task-set derivation. The latency is shown next to `IDP/CDP Latency` field. The throughput is the reciprocal of the output actor(s) period(s).

Platform and Mapping Specifications
-----------------------------------
The platform and mapping specifications can be found under different scheduling algorithms. The supported mapping are listed in the supported_mappings variable in PlatformGenerator.py which determins the platform and mapping specifications. It can be invoked as follows:

    > ./PlatformGenerator.py output_dir time_in_cycles alloc_alg app_1_factor app_1_tok_size app_1.csf app_2_factor app_2_tok_size app_2.csf ...

where:

- `output_dir`: is the directory where the output will be produced
- `time_in_cycles`: is a binary flag (1 or 0) which specfifies whether the time will be in clock cycles or OS clock ticks
- `alloc_alg`: is the allocation algorithm type
- `app_i_factor`: is the period scaling factor
- `app_i_tok_size`: is the maximum token size for app_i
- `app_i.csf`: is the path to the .csf file containing app_i

`PlatformGenerator.py` produces two files: platform specifications (`platform.pla`) and mapping specifications (`mapping.map`)


Graph Formats
-------------

csdf-rtschedtools accepts as an input the graphs in StreamIt Modified Format (GPH Format) produced by [pntools](https://daedalus.liacs.nl/pntools/). An example of the GPH format for an SDF graph consisting of two actors (producer and consumer) is shown below:

    node_number:2
    node:
	    id:0
	    name:ND_0
	    function:producer
	    length:1
	    wcet:5
	    port_number:1
	    port:
		    type:out
		    id:1
		    rate:1
    node:
	    id:1
	    name:ND_1
	    function:consumer
	    length:1
	    wcet:2
	    port_number:1
	    port:
		    type:in
		    id:1
		    rate:1
    edge_number:1
    edge:
        id:0
        name:ED_0
        src:0 1
        dst:1 1	    

In order to facilitate conversion between formats, we provide a conversion tools called `CSDFConverter.py`. It is capable of accepting a .gph file as input and producing either [SDF For Free](http://www.es.ele.tue.nl/sdf3/) XML format or Compact StreamIt Format (CSF). 

Bugs/Questions
---------------
In case of bugs/questions, please contact [Mohamed Bamakhrama](mohamed@liacs.nl)

LICENSE
-------
This tool is released under the BSD license shown in LICENSE.md
