#!/usr/bin/python

__author__ =  'Mohamed A. Bamakhrama'
__description__ = 'Text Handler'

import os.path

import Utilities

class TextHandler:
	"""Text Handler"""
	
	def __init__(self, pg_instance, output_dir, partitions):
		self.pg_instance = pg_instance
		self.output_dir = output_dir
		self.partitions = partitions
		self.num_of_apps = self.pg_instance.get_num_of_graphs()

	def generate_mapping_file(self):
		mapping = os.path.join(self.output_dir, "mapping.txt")
		fp = open(mapping, 'w')
		tn = 0
		
		partitions = self.partitions
		
		fp.write("number_of_processors:" + str(len(partitions)) + "\n")

		proc_cnt = 0
		task_cnt = 0
		for p in partitions:
			fp.write(Utilities.tabs(tn) + "processor:\n")
			tn += 1
			fp.write(Utilities.tabs(tn) + "id:" + str(proc_cnt) + "\n")
			u_sum = 0.0
			for t in p:
				u_sum += t.get_utilization()
			fp.write(Utilities.tabs(tn) + "utilization:" + str(u_sum) + "\n")
			for t in p:
				fp.write(Utilities.tabs(tn) + "task:\n")
				tn += 1
				if self.num_of_apps > 1:
					fp.write(Utilities.tabs(tn) + "name:" + t.get_full_name() + "\n")
				else:
					fp.write(Utilities.tabs(tn) + "name:" + t.get_name() + "\n")
				fp.write(Utilities.tabs(tn) + "function:" + t.get_function() + "\n")
				fp.write(Utilities.tabs(tn) + "wcet:" + str(t.get_wcet()) + "\n")
				fp.write(Utilities.tabs(tn) + "period:" + str(t.get_period()) + "\n")
				fp.write(Utilities.tabs(tn) + "deadline:" + str(t.get_deadline()) + "\n")
				fp.write(Utilities.tabs(tn) + "start_time:" + str(t.get_start_time()) + "\n")
				fp.write(Utilities.tabs(tn) + "utilization:" + str(t.get_utilization()) + "\n")
				fp.write(Utilities.tabs(tn) + "priority:" + str(t.get_priority()) + "\n")
				tn -= 1
				task_cnt += 1
			tn -= 1
			proc_cnt += 1
		tn -= 1
		
		num_of_chan = 0
		for g in self.pg_instance.get_graphs():
			num_of_chan += len(g.get_channels())
		
		fp.write(Utilities.tabs(tn) + "number_of_channels:" + str(num_of_chan) + "\n")
		tn += 1
		
		chan_cnt = 0
		for a in self.pg_instance.get_tasks():
			actor = a
			for outport in actor.get_outports():
				ch = actor.get_graph().get_channel(outport)
				successor = actor.get_graph().get_actor(ch.get_destination())
				fp.write(Utilities.tabs(tn) + "channel:\n")
				tn += 1
				fp.write(Utilities.tabs(tn) + "id:" + str(chan_cnt) + "\n")
				fp.write(Utilities.tabs(tn) + "size:" + str(ch.get_buffer_size()) + "\n")
				if self.num_of_apps > 1:
					fp.write(Utilities.tabs(tn) + "source_task:" + actor.get_full_name() + "\n")
					fp.write(Utilities.tabs(tn) + "destination_task:" + successor.get_full_name()  + "\n")
				else:
					fp.write(Utilities.tabs(tn) + "source_task:" + actor.get_name() + "\n")
					fp.write(Utilities.tabs(tn) + "destination_task:" + successor.get_name()  + "\n")
					
				fp.write(Utilities.tabs(tn) + "source_processor:" + str(actor.get_affinity()) + " \n")
				fp.write(Utilities.tabs(tn) + "destination_processor:" + str(successor.get_affinity()) + " \n")
				
				prd_tokens = sum(ch.get_production_sequence())
				prd_len = len(ch.get_production_sequence())
				cns_tokens = sum(ch.get_consumption_sequence())
				cns_len = len(ch.get_consumption_sequence())
				
				fp.write(Utilities.tabs(tn) + "write_rate:" + str(prd_tokens) + "/" + str(prd_len * actor.get_period()) + "\n")
				fp.write(Utilities.tabs(tn) + "read_rate:" + str(cns_tokens) + "/" + str(cns_len * successor.get_period()) + "\n")						
				
				tn -= 1
				chan_cnt += 1
		fp.flush()
		fp.close()

